@@ -17,11 +17,12 @@ from tacotron.checkpoint_handling import (CheckpointDict, create, get_hparams, g
                                           get_model_state, get_optimizer_state, get_scheduler_state,
                                           get_speaker_embedding_weights, get_speaker_mapping,
                                           get_stress_mapping, get_symbol_embedding_weights,
-                                          get_symbol_mapping)
+                                          get_symbol_mapping, get_tone_mapping)
 from tacotron.dataloader import (SymbolsMelCollate, create_speaker_mapping,
-                                 create_symbol_and_stress_mapping, create_symbol_mapping,
-                                 get_speaker_mappings_count, get_stress_mappings_count,
-                                 get_symbol_mappings_count, parse_batch, prepare_trainloader,
+                                 create_symbol_and_stress_mapping, create_symbol_and_tone_mapping,
+                                 create_symbol_mapping, get_speaker_mappings_count,
+                                 get_stress_mappings_count, get_symbol_mappings_count,
+                                 get_tone_mappings_count, parse_batch, prepare_trainloader,
                                  prepare_valloader)
 from tacotron.hparams import ExperimentHParams, HParams, OptimizerHParams
 from tacotron.logger import Tacotron2Logger
@@ -115,7 +116,7 @@ def log_symbol_weights(model: Tacotron2, logger: Logger) -> None:
   logger.info(str(model.state_dict()[SYMBOL_EMBEDDING_LAYER_NAME]))
 
 
-def start_training(custom_hparams: Optional[Dict[str, str]], taco_logger: Tacotron2Logger, trainset: Entries, valset: Entries, save_callback: Callable[[CheckpointDict], None], checkpoint: Optional[CheckpointDict], pretrained_model: Optional[CheckpointDict], warm_start: bool, map_symbol_weights: bool, custom_symbol_weights_map: Optional[SymbolToSymbolMapping], map_speaker_weights: bool, map_from_speaker_name: Optional[str], device: torch.device, logger: Logger, checkpoint_logger: Logger) -> None:
+def start_training(custom_hparams: Optional[Dict[str, str]], taco_logger: Tacotron2Logger, trainset: Entries, valset: Entries, save_callback: Callable[[CheckpointDict], None], checkpoint: Optional[CheckpointDict], pretrained_model: Optional[CheckpointDict], warm_start: bool, map_symbol_weights: bool, custom_symbol_weights_map: Optional[SymbolToSymbolMapping], map_speaker_weights: bool, map_from_speaker_name: Optional[str], device: torch.device, logger: Logger, checkpoint_logger: Logger) -> bool:
   complete_start = time.time()
 
   if checkpoint is not None:
@@ -130,8 +131,11 @@ def start_training(custom_hparams: Optional[Dict[str, str]], taco_logger: Tacotr
   init_torch(hparams)
 
   n_stresses = None
+  n_tones = None
   stress_mapping = None
+  tone_mapping = None
   if hparams.use_stress_embedding:
+    assert not hparams.use_tone_embedding
     if checkpoint is None:
       symbol_mapping, stress_mapping = create_symbol_and_stress_mapping(
           valset, trainset, hparams.symbols_are_ipa)
@@ -139,6 +143,16 @@ def start_training(custom_hparams: Optional[Dict[str, str]], taco_logger: Tacotr
       symbol_mapping = get_symbol_mapping(checkpoint)
       stress_mapping = get_stress_mapping(checkpoint)
     n_stresses = get_stress_mappings_count(stress_mapping)
+  elif hparams.use_tone_embedding:
+    assert not hparams.use_stress_embedding
+    assert hparams.symbols_are_ipa
+    if checkpoint is None:
+      symbol_mapping, tone_mapping = create_symbol_and_tone_mapping(
+          valset, trainset)
+    else:
+      symbol_mapping = get_symbol_mapping(checkpoint)
+      tone_mapping = get_tone_mapping(checkpoint)
+    n_tones = get_tone_mappings_count(tone_mapping)
   else:
     if checkpoint is None:
       symbol_mapping = create_symbol_mapping(valset, trainset)
@@ -161,8 +175,18 @@ def start_training(custom_hparams: Optional[Dict[str, str]], taco_logger: Tacotr
   if hparams.use_stress_embedding:
     logger.info(
         f"Stresses: {' '.join(stress_mapping.keys())} (#{len(stress_mapping)})")
+    if len(stress_mapping) != 4:
+      logger.error("Not all stress marks exist in the data!")
+      return False
   else:
     logger.info("Use no stress embedding.")
+
+  if hparams.use_tone_embedding:
+    logger.info(
+        f"Tones: {' '.join(tone_mapping.keys())} (#{len(tone_mapping)})")
+  else:
+    logger.info("Use no tone embedding.")
+
   if hparams.use_speaker_embedding:
     logger.info(
         f"Speakers: {', '.join(sorted(speaker_mapping.keys()))} (#{len(speaker_mapping)}, dim: {hparams.speakers_embedding_dim})")
@@ -175,6 +199,7 @@ def start_training(custom_hparams: Optional[Dict[str, str]], taco_logger: Tacotr
       n_symbols=n_symbols,
       n_stresses=n_stresses,
       n_speakers=n_speakers,
+      n_tones=n_tones,
   )
 
   model = cast(Tacotron2, try_copy_to(model, device))
@@ -206,12 +231,12 @@ def start_training(custom_hparams: Optional[Dict[str, str]], taco_logger: Tacotr
       if pretrained_model is None:
         logger.error(
             "Warm start: For warm start a pretrained model must be provided!")
-        return
+        return False
 
       success = warm_start_model(
           model, pretrained_model, hparams, logger)
       if not success:
-        return
+        return False
     else:
       logger.info("Didn't used warm start.")
 
@@ -220,19 +245,31 @@ def start_training(custom_hparams: Optional[Dict[str, str]], taco_logger: Tacotr
       if pretrained_model is None:
         logger.error(
             "Mapping symbol weights: For mapping symbol weights a pretrained model must be provided!")
-        return
+        return False
 
       pre_hparams = get_hparams(pretrained_model)
       if pre_hparams.stress_embedding_dim != hparams.stress_embedding_dim:
         logger.error(
             "Mapping symbol weights: Stress embedding dimensions do not match!")
-        return
+        return False
+
+      pre_hparams = get_hparams(pretrained_model)
+      if pre_hparams.tone_embedding_dim != hparams.tone_embedding_dim:
+        logger.error(
+            "Mapping symbol weights: Tone embedding dimensions do not match!")
+        return False
 
       pre_stress_mapping = get_stress_mapping(pretrained_model)
       if pre_stress_mapping.keys() != stress_mapping.keys():
         logger.error(
             f"Mapping symbol weights: Stress mappings are not equal! '{' '.join(pre_stress_mapping.keys())}' vs. '{' '.join(stress_mapping.keys())}'")
-        return
+        return False
+
+      pre_tone_mapping = get_tone_mapping(pretrained_model)
+      if pre_tone_mapping.keys() != tone_mapping.keys():
+        logger.error(
+            f"Mapping symbol weights: Tone mappings are not equal! '{' '.join(pre_tone_mapping.keys())}' vs. '{' '.join(tone_mapping.keys())}'")
+        return False
 
       pre_symbol_weights = get_symbol_embedding_weights(pretrained_model)
       pre_symbol_mapping = get_symbol_mapping(pretrained_model)
@@ -302,25 +339,25 @@ def start_training(custom_hparams: Optional[Dict[str, str]], taco_logger: Tacotr
       if pretrained_model is None:
         logger.error(
             "Mapping speaker weights: For mapping speaker weights a pretrained model must be provided!")
-        return
+        return False
 
       if map_from_speaker_name is None:
         logger.error(
             "Mapping speaker weights: A speaker name is required for mapping speaker weights.")
-        return
+        return False
 
       pre_hparams = get_hparams(pretrained_model)
       both_models_use_speaker_emb = hparams.use_speaker_embedding and pre_hparams.use_speaker_embedding
       if not both_models_use_speaker_emb:
         logger.error(
             "Mapping speaker weights: Couldn't map speaker embeddings because one of the models use no speaker embeddings!")
-        return
+        return False
 
       pre_speaker_mapping = get_speaker_mapping(pretrained_model)
       if map_from_speaker_name not in pre_speaker_mapping:
         logger.error(
             f"Mapping speaker weights: Speaker '{map_from_speaker_name}' was not found in weights checkpoint.")
-        return
+        return False
 
       pre_speaker_embedding = get_speaker_embedding_weights(
           pretrained_model)
@@ -345,31 +382,32 @@ def start_training(custom_hparams: Optional[Dict[str, str]], taco_logger: Tacotr
   # log_symbol_weights(model, logger)
 
   collate_fn = SymbolsMelCollate(
-      n_frames_per_step=hparams.n_frames_per_step,
-      use_stress=hparams.use_stress_embedding,
-      use_speakers=hparams.use_speaker_embedding,
+    n_frames_per_step=hparams.n_frames_per_step,
+    use_stress=hparams.use_stress_embedding,
+    use_tones=hparams.use_tone_embedding,
+    use_speakers=hparams.use_speaker_embedding,
   )
 
   val_loader = prepare_valloader(hparams, collate_fn, valset,
-                                 symbol_mapping, stress_mapping, speaker_mapping, device, logger)
+                                 symbol_mapping, stress_mapping, tone_mapping, speaker_mapping, device, logger)
   train_loader = prepare_trainloader(
-      hparams, collate_fn, trainset, symbol_mapping, stress_mapping, speaker_mapping, device, logger)
+      hparams, collate_fn, trainset, symbol_mapping, stress_mapping, tone_mapping, speaker_mapping, device, logger)
 
   batch_iterations = len(train_loader)
   enough_traindata = batch_iterations > 0
   if not enough_traindata:
     msg = "Not enough training data!"
     logger.error(msg)
-    return
+    return False
 
   save_it_settings = SaveIterationSettings(
-      epochs=hparams.epochs,
-      iterations=hparams.iterations,
-      batch_iterations=batch_iterations,
-      save_first_iteration=hparams.save_first_iteration,
-      save_last_iteration=True,
-      iters_per_checkpoint=hparams.iters_per_checkpoint,
-      epochs_per_checkpoint=hparams.epochs_per_checkpoint
+    epochs=hparams.epochs,
+    iterations=hparams.iterations,
+    batch_iterations=batch_iterations,
+    save_first_iteration=hparams.save_first_iteration,
+    save_last_iteration=True,
+    iters_per_checkpoint=hparams.iters_per_checkpoint,
+    epochs_per_checkpoint=hparams.epochs_per_checkpoint
   )
 
   last_iteration = get_last_iteration(
@@ -521,15 +559,16 @@ def start_training(custom_hparams: Optional[Dict[str, str]], taco_logger: Tacotr
       save_it = check_save_it(epoch, iteration, save_it_settings)
       if save_it:
         checkpoint = create(
-            model=model,
-            optimizer=optimizer,
-            hparams=hparams,
-            iteration=iteration,
-            speaker_mapping=speaker_mapping,
-            learning_rate=get_lr(optimizer),
-            stress_mapping=stress_mapping,
-            symbol_mapping=symbol_mapping,
-            scheduler=scheduler,
+          model=model,
+          optimizer=optimizer,
+          hparams=hparams,
+          iteration=iteration,
+          speaker_mapping=speaker_mapping,
+          learning_rate=get_lr(optimizer),
+          stress_mapping=stress_mapping,
+          symbol_mapping=symbol_mapping,
+          tone_mapping=tone_mapping,
+          scheduler=scheduler,
         )
 
         save_callback(checkpoint)
@@ -557,6 +596,7 @@ def start_training(custom_hparams: Optional[Dict[str, str]], taco_logger: Tacotr
 
   duration_s = time.time() - complete_start
   logger.info(f'Finished training. Total duration: {duration_s / 60:.2f}m')
+  return True
 
 
 def adjust_lr(hparams: HParams, optimizer: Optimizer, epoch: int, scheduler, logger: Logger) -> None:
@@ -594,8 +634,8 @@ def set_lr(optimizer: Optimizer, lr: float) -> None:
     g['lr'] = lr
 
 
-def load_model(hparams: HParams, checkpoint: Optional[CheckpointDict], n_symbols: int, n_stresses: Optional[int], n_speakers: Optional[int]) -> Tacotron2:
-  model = Tacotron2(hparams, n_symbols, n_stresses, n_speakers)
+def load_model(hparams: HParams, checkpoint: Optional[CheckpointDict], n_symbols: int, n_stresses: Optional[int], n_speakers: Optional[int], n_tones: Optional[int]) -> Tacotron2:
+  model = Tacotron2(hparams, n_symbols, n_stresses, n_speakers, n_tones)
   if checkpoint is not None:
     model_state = get_model_state(checkpoint)
     model.load_state_dict(model_state)
@@ -632,7 +672,8 @@ def load_scheduler(optimizer: Adam, hparams: OptimizerHParams, checkpoint: Optio
     scheduler = ExponentialLR(
         optimizer=optimizer,
         gamma=scheduler_state["gamma"],
-        last_epoch=scheduler_state["last_epoch"],
+        # -1 because there is probably a bug in scheduler: it adjusts the learning rate when starting training if its scheduler_state[last_epoch]
+        last_epoch=-1,
         verbose=scheduler_state["verbose"],
     )
     scheduler.load_state_dict(scheduler_state)
@@ -661,6 +702,11 @@ def warm_start_model(model: Tacotron2, warm_model: CheckpointDict, hparams: HPar
     logger.error(msg)
     return False
 
+  if hparams.use_tone_embedding and not warm_model_hparams.use_tone_embedding:
+    msg = "Warm start: Warm model did not used a tone embedding!"
+    logger.error(msg)
+    return False
+
   if hparams.use_speaker_embedding:
     if not warm_model_hparams.use_speaker_embedding:
       msg = "Warm start: Warm model did not used a speaker embedding!"
